import pytest

from pykka import ActorRegistry

pytestmark = pytest.mark.usefixtures("stop_all")


class ActorBase:
    received_messages = None

    def __init__(self):
        super().__init__()
        self.received_messages = []

    def on_receive(self, message):
        self.received_messages.append(message)


@pytest.fixture(scope="module")
def actor_a_class(runtime):
    class ActorA(ActorBase, runtime.actor_class):
        pass

    return ActorA


@pytest.fixture(scope="module")
def actor_b_class(runtime):
    class ActorB(ActorBase, runtime.actor_class):
        pass

    return ActorB


@pytest.fixture
def actor_ref(actor_a_class):
    return actor_a_class.start()


@pytest.fixture
def a_actor_refs(actor_a_class):
    return [actor_a_class.start() for _ in range(3)]


@pytest.fixture
def b_actor_refs(actor_b_class):
    return [actor_b_class.start() for _ in range(5)]


def test_actor_is_registered_when_started(actor_ref):
    assert actor_ref in ActorRegistry.get_all()


def test_actor_is_unregistered_when_stopped(actor_ref):
    assert actor_ref in ActorRegistry.get_all()

    actor_ref.stop()

    assert actor_ref not in ActorRegistry.get_all()


def test_actor_may_be_registered_manually(actor_ref):
    ActorRegistry.unregister(actor_ref)
    assert actor_ref not in ActorRegistry.get_all()

    ActorRegistry.register(actor_ref)

    assert actor_ref in ActorRegistry.get_all()


def test_actor_may_be_unregistered_multiple_times_without_error(actor_ref):
    ActorRegistry.unregister(actor_ref)
    assert actor_ref not in ActorRegistry.get_all()

    ActorRegistry.unregister(actor_ref)
    assert actor_ref not in ActorRegistry.get_all()

    ActorRegistry.register(actor_ref)
    assert actor_ref in ActorRegistry.get_all()


def test_all_actors_can_be_stopped_through_registry(a_actor_refs, b_actor_refs):
    assert len(ActorRegistry.get_all()) == 8

    ActorRegistry.stop_all(block=True)

    assert len(ActorRegistry.get_all()) == 0


def test_stop_all_stops_last_started_actor_first_if_blocking(mocker):
    mocker.patch.object(ActorRegistry, "get_all")

    stopped_actors = []
    started_actors = [mocker.Mock(name=i) for i in range(3)]
    started_actors[0].stop.side_effect = lambda *a, **kw: stopped_actors.append(
        started_actors[0]
    )
    started_actors[1].stop.side_effect = lambda *a, **kw: stopped_actors.append(
        started_actors[1]
    )
    started_actors[2].stop.side_effect = lambda *a, **kw: stopped_actors.append(
        started_actors[2]
    )
    ActorRegistry.get_all.return_value = started_actors

    ActorRegistry.stop_all(block=True)

    assert stopped_actors[0] == started_actors[2]
    assert stopped_actors[1] == started_actors[1]
    assert stopped_actors[2] == started_actors[0]


def test_actors_may_be_looked_up_by_class(actor_a_class, a_actor_refs, b_actor_refs):
    result = ActorRegistry.get_by_class(actor_a_class)

    for a_actor in a_actor_refs:
        assert a_actor in result
    for b_actor in b_actor_refs:
        assert b_actor not in result


def test_actors_may_be_looked_up_by_superclass(
    actor_a_class, a_actor_refs, b_actor_refs
):
    result = ActorRegistry.get_by_class(actor_a_class)

    for a_actor in a_actor_refs:
        assert a_actor in result
    for b_actor in b_actor_refs:
        assert b_actor not in result


def test_actors_may_be_looked_up_by_class_name(
    actor_a_class, a_actor_refs, b_actor_refs
):
    result = ActorRegistry.get_by_class_name("ActorA")

    for a_actor in a_actor_refs:
        assert a_actor in result
    for b_actor in b_actor_refs:
        assert b_actor not in result


def test_actors_may_be_looked_up_by_urn(actor_ref):
    result = ActorRegistry.get_by_urn(actor_ref.actor_urn)

    assert result == actor_ref


def test_get_by_urn_returns_none_if_not_found():
    result = ActorRegistry.get_by_urn("urn:foo:bar")

    assert result is None


def test_broadcast_sends_message_to_all_actors_if_no_target(a_actor_refs, b_actor_refs):
    ActorRegistry.broadcast({"command": "foo"})

    running_actors = ActorRegistry.get_all()
    assert running_actors

    for actor_ref in running_actors:
        received_messages = actor_ref.proxy().received_messages.get()
        assert {"command": "foo"} in received_messages


def test_broadcast_sends_message_to_all_actors_of_given_class(
    actor_a_class, actor_b_class
):
    ActorRegistry.broadcast({"command": "foo"}, target_class=actor_a_class)

    for actor_ref in ActorRegistry.get_by_class(actor_a_class):
        received_messages = actor_ref.proxy().received_messages.get()
        assert {"command": "foo"} in received_messages

    for actor_ref in ActorRegistry.get_by_class(actor_b_class):
        received_messages = actor_ref.proxy().received_messages.get()
        assert {"command": "foo"} not in received_messages


def test_broadcast_sends_message_to_all_actors_of_given_class_name(
    actor_a_class, actor_b_class
):
    ActorRegistry.broadcast({"command": "foo"}, target_class="ActorA")

    for actor_ref in ActorRegistry.get_by_class(actor_a_class):
        received_messages = actor_ref.proxy().received_messages.get()
        assert {"command": "foo"} in received_messages

    for actor_ref in ActorRegistry.get_by_class(actor_b_class):
        received_messages = actor_ref.proxy().received_messages.get()
        assert {"command": "foo"} not in received_messages



import logging
import threading

logger = logging.getLogger("pykka")


__all__ = ["ActorRegistry"]


class ActorRegistry:
    """
    Registry which provides easy access to all running actors.

    Contains global state, but should be thread-safe.
    """

    _actor_refs = []
    _actor_refs_lock = threading.RLock()

    @classmethod
    def broadcast(cls, message, target_class=None):
        """
        Broadcast ``message`` to all actors of the specified ``target_class``.

        If no ``target_class`` is specified, the message is broadcasted to all
        actors.

        :param message: the message to send
        :type message: any

        :param target_class: optional actor class to broadcast the message to
        :type target_class: class or class name
        """
        if isinstance(target_class, str):
            targets = cls.get_by_class_name(target_class)
        elif target_class is not None:
            targets = cls.get_by_class(target_class)
        else:
            targets = cls.get_all()
        for ref in targets:
            ref.tell(message)

    @classmethod
    def get_all(cls):
        """
        Get :class:`ActorRef <pykka.ActorRef>` for all running actors.

        :returns: list of :class:`pykka.ActorRef`
        """
        with cls._actor_refs_lock:
            return cls._actor_refs[:]

    @classmethod
    def get_by_class(cls, actor_class):
        """
        Get :class:`ActorRef` for all running actors of the given class, or of
        any subclass of the given class.

        :param actor_class: actor class, or any superclass of the actor
        :type actor_class: class

        :returns: list of :class:`pykka.ActorRef`
        """
        with cls._actor_refs_lock:
            return [
                ref
                for ref in cls._actor_refs
                if issubclass(ref.actor_class, actor_class)
            ]

    @classmethod
    def get_by_class_name(cls, actor_class_name):
        """
        Get :class:`ActorRef` for all running actors of the given class
        name.

        :param actor_class_name: actor class name
        :type actor_class_name: string

        :returns: list of :class:`pykka.ActorRef`
        """
        with cls._actor_refs_lock:
            return [
                ref
                for ref in cls._actor_refs
                if ref.actor_class.__name__ == actor_class_name
            ]

    @classmethod
    def get_by_urn(cls, actor_urn):
        """
        Get an actor by its universally unique URN.

        :param actor_urn: actor URN
        :type actor_urn: string

        :returns: :class:`pykka.ActorRef` or :class:`None` if not found
        """
        with cls._actor_refs_lock:
            refs = [ref for ref in cls._actor_refs if ref.actor_urn == actor_urn]
            if refs:
                return refs[0]

    @classmethod
    def register(cls, actor_ref):
        """
        Register an :class:`ActorRef` in the registry.

        This is done automatically when an actor is started, e.g. by calling
        :meth:`Actor.start() <pykka.Actor.start>`.

        :param actor_ref: reference to the actor to register
        :type actor_ref: :class:`pykka.ActorRef`
        """
        with cls._actor_refs_lock:
            cls._actor_refs.append(actor_ref)
        logger.debug(f"Registered {actor_ref}")

    @classmethod
    def stop_all(cls, block=True, timeout=None):
        """
        Stop all running actors.

        ``block`` and ``timeout`` works as for
        :meth:`ActorRef.stop() <pykka.ActorRef.stop>`.

        If ``block`` is :class:`True`, the actors are guaranteed to be stopped
        in the reverse of the order they were started in. This is helpful if
        you have simple dependencies in between your actors, where it is
        sufficient to shut down actors in a LIFO manner: last started, first
        stopped.

        If you have more complex dependencies in between your actors, you
        should take care to shut them down in the required order yourself, e.g.
        by stopping dependees from a dependency's
        :meth:`on_stop() <pykka.Actor.on_stop>` method.

        :returns: If not blocking, a list with a future for each stop action.
            If blocking, a list of return values from
            :meth:`pykka.ActorRef.stop`.
        """
        return [ref.stop(block, timeout) for ref in reversed(cls.get_all())]

    @classmethod
    def unregister(cls, actor_ref):
        """
        Remove an :class:`ActorRef <pykka.ActorRef>` from the registry.

        This is done automatically when an actor is stopped, e.g. by calling
        :meth:`Actor.stop() <pykka.Actor.stop>`.

        :param actor_ref: reference to the actor to unregister
        :type actor_ref: :class:`pykka.ActorRef`
        """
        removed = False
        with cls._actor_refs_lock:
            if actor_ref in cls._actor_refs:
                cls._actor_refs.remove(actor_ref)
                removed = True
        if removed:
            logger.debug(f"Unregistered {actor_ref}")
        else:
            logger.debug(f"Unregistered {actor_ref} (not found in registry)")



"""Pykka documentation build configuration file"""

import toml

project = "Pykka"
author = "Stein Magnus Jodal and contributors"
copyright = f"2010-2022, {author}"

release = toml.load("../pyproject.toml")["tool"]["poetry"]["version"]
version = ".".join(release.split(".")[:2])

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.extlinks",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
]

html_theme = "sphinx_rtd_theme"
html_use_modindex = True
html_use_index = True
html_split_index = False
html_show_sourcelink = True
modindex_common_prefix = ["pykka."]

autodoc_member_order = "bysource"

extlinks = {"issue": ("https://github.com/jodal/pykka/issues/%s", "#")}

intersphinx_mapping = {"python": ("https://docs.python.org/3", None)}



from collections.abc import Callable, Generator, Iterable
from typing import Any, Generic, TypeVar

from pykka._types import OptExcInfo

_T = TypeVar("_T")
J = TypeVar("J")  # For when T is Iterable[J]  # noqa

_M = TypeVar("_M")  # For Future.map()
_R = TypeVar("_R")  # For Future.reduce()

GetHookFunc = Callable[[float | None], _T]

class Future(Generic[_T]):
    _get_hook: GetHookFunc | None
    _get_hook_result: _T | None
    def get(self, timeout: float | None = ...) -> _T: ...
    def set(self, value: _T | None = ...) -> None: ...
    def set_exception(self, exc_info: OptExcInfo | None = ...) -> None: ...
    def set_get_hook(self, func: GetHookFunc) -> None: ...
    def filter(
        self: Future[Iterable[J]], func: Callable[[J], bool]
    ) -> Future[Iterable[J]]: ...
    def join(self, *futures: Future[Any]) -> Future[Iterable[Any]]: ...
    def map(self, func: Callable[[_T], _M]) -> Future[_M]: ...
    def reduce(
        self: Future[Iterable[J]], func: Callable[[_R, J], _R], *args: _R
    ) -> Future[_R]: ...
    def __await__(self) -> Generator[None, None, _T]: ...

def get_all(
    futures: Iterable[Future[Any]], timeout: float | None = ...
) -> Iterable[Any]: ...



from collections.abc import Callable

import pytest


@pytest.fixture
def actor_class(runtime):
    class ActorForMocking(runtime.actor_class):
        _a_rw_property = "a_rw_property"

        @property
        def a_rw_property(self):
            return self._a_rw_property

        @a_rw_property.setter
        def a_rw_property(self, value):
            self._a_rw_property = value

        def a_method(self):
            raise Exception("This method should be mocked")

    return ActorForMocking


@pytest.fixture
def proxy(actor_class):
    proxy = actor_class.start().proxy()
    yield proxy
    proxy.stop()


def test_actor_with_noncallable_mock_property_works(actor_class, stop_all, mocker):
    mock = mocker.NonCallableMock()
    mock.__get__ = mocker.Mock(return_value="mocked property value")
    assert not isinstance(mock, Callable)

    actor_class.a_rw_property = mock
    proxy = actor_class.start().proxy()

    # When using NonCallableMock to fake the property, the value still behaves
    # as a property when access through the proxy.
    assert proxy.a_rw_property.get() == "mocked property value"

    assert mock.__get__.call_count == 1


def test_actor_with_callable_mock_property_does_not_work(actor_class, stop_all, mocker):
    mock = mocker.Mock()
    mock.__get__ = mocker.Mock(return_value="mocked property value")
    assert isinstance(mock, Callable)

    actor_class.a_rw_property = mock
    proxy = actor_class.start().proxy()

    # XXX Because Mock and MagicMock are callable by default, they cause the
    # property to be wrapped in a `CallableProxy`. Thus, the property no
    # longer behaves as a property when mocked and accessed through a proxy.
    with pytest.raises(AttributeError) as exc_info:
        assert proxy.a_rw_property.get()

    assert "'CallableProxy' object has no attribute 'get'" in str(exc_info.value)


def test_actor_with_mocked_method_works(actor_class, stop_all, mocker):
    mock = mocker.MagicMock(return_value="mocked method return")
    mocker.patch.object(actor_class, "a_method", new=mock)
    proxy = actor_class.start().proxy()

    assert proxy.a_method().get() == "mocked method return"

    assert mock.call_count == 1



import logging
import threading
import time
from collections import namedtuple

import pytest

from pykka import ActorRegistry, ThreadingActor, ThreadingFuture
from tests.log_handler import PykkaTestLogHandler

Runtime = namedtuple(
    "Runtime",
    ["name", "actor_class", "event_class", "future_class", "sleep_func"],
)


RUNTIMES = {
    "threading": pytest.param(
        Runtime(
            name="threading",
            actor_class=ThreadingActor,
            event_class=threading.Event,
            future_class=ThreadingFuture,
            sleep_func=time.sleep,
        ),
        id="threading",
    )
}


@pytest.fixture(scope="session", params=RUNTIMES.values())
def runtime(request):
    return request.param


@pytest.fixture
def stop_all():
    yield
    ActorRegistry.stop_all()


@pytest.fixture
def log_handler():
    log_handler = PykkaTestLogHandler()

    root_logger = logging.getLogger()
    root_logger.addHandler(log_handler)
    # pytest sets the root logger level to WARNING. We reset it to NOTSET
    # so that all log messages reaches our log handler.
    root_logger.setLevel(logging.NOTSET)

    yield log_handler

    log_handler.close()


@pytest.fixture
def events(runtime):
    class Events:
        on_start_was_called = runtime.event_class()
        on_stop_was_called = runtime.event_class()
        on_failure_was_called = runtime.event_class()
        greetings_was_received = runtime.event_class()
        actor_registered_before_on_start_was_called = runtime.event_class()

    return Events()


@pytest.fixture(scope="module")
def early_failing_actor_class(runtime):
    class EarlyFailingActor(runtime.actor_class):
        def __init__(self, events):
            super().__init__()
            self.events = events

        def on_start(self):
            try:
                raise RuntimeError("on_start failure")
            finally:
                self.events.on_start_was_called.set()

    return EarlyFailingActor


@pytest.fixture(scope="module")
def late_failing_actor_class(runtime):
    class LateFailingActor(runtime.actor_class):
        def __init__(self, events):
            super().__init__()
            self.events = events

        def on_start(self):
            self.stop()

        def on_stop(self):
            try:
                raise RuntimeError("on_stop failure")
            finally:
                self.events.on_stop_was_called.set()

    return LateFailingActor


@pytest.fixture(scope="module")
def failing_on_failure_actor_class(runtime):
    class FailingOnFailureActor(runtime.actor_class):
        def __init__(self, events):
            super().__init__()
            self.events = events

        def on_receive(self, message):
            if message.get("command") == "raise exception":
                raise Exception("on_receive failure")
            else:
                super().on_receive(message)

        def on_failure(self, *args):
            try:
                raise RuntimeError("on_failure failure")
            finally:
                self.events.on_failure_was_called.set()

    return FailingOnFailureActor


@pytest.fixture
def future(runtime):
    return runtime.future_class()


@pytest.fixture
def futures(runtime):
    return [runtime.future_class() for _ in range(3)]



import pytest

import pykka


class NestedWithNoMarker:
    inner = "nested_with_no_marker.inner"


class NestedWithNoMarkerAndSlots:
    __slots__ = ["inner"]

    def __init__(self):
        self.inner = "nested_with_no_marker_and_slots.inner"


@pykka.traversable
class NestedWithDecoratorMarker:
    inner = "nested_with_decorator_marker.inner"


class NestedWithAttrMarker:
    pykka_traversable = True
    inner = "nested_with_attr_marker.inner"


class NestedWithAttrMarkerAndSlots:
    __slots__ = ["pykka_traversable", "inner"]

    def __init__(self):
        # Objects using '__slots__' cannot have class attributes.
        self.pykka_traversable = True
        self.inner = "nested_with_attr_marker_and_slots.inner"


@pytest.fixture
def actor_class(runtime):
    class ActorWithTraversableObjects(runtime.actor_class):
        nested_with_no_marker = NestedWithNoMarker()
        nested_with_function_marker = pykka.traversable(NestedWithNoMarker())
        nested_with_decorator_marker = NestedWithDecoratorMarker()
        nested_with_attr_marker = NestedWithAttrMarker()
        nested_with_attr_marker_and_slots = NestedWithAttrMarkerAndSlots()

        @property
        def nested_object_property(self):
            return NestedWithAttrMarker()

    return ActorWithTraversableObjects


@pytest.fixture
def proxy(actor_class):
    proxy = actor_class.start().proxy()
    yield proxy
    proxy.stop()


def test_attr_without_marker_cannot_be_traversed(proxy):
    with pytest.raises(AttributeError) as exc_info:
        proxy.nested_with_no_marker.inner.get()

    assert "object has no attribute 'inner'" in str(exc_info.value)


@pytest.mark.parametrize(
    "attr_name, expected",
    [
        ("nested_with_function_marker", "nested_with_no_marker.inner"),
        ("nested_with_decorator_marker", "nested_with_decorator_marker.inner"),
        ("nested_with_attr_marker", "nested_with_attr_marker.inner"),
        (
            "nested_with_attr_marker_and_slots",
            "nested_with_attr_marker_and_slots.inner",
        ),
    ],
)
def test_attr_of_traversable_attr_can_be_read(proxy, attr_name, expected):
    attr = getattr(proxy, attr_name)

    assert attr.inner.get() == expected


def test_traversable_object_returned_from_property_is_not_traversed(proxy):
    # In Pykka < 2, it worked like this:
    # assert proxy.nested_object_property.inner.get() == 'nested.inner'

    # In Pykka >= 2, the property getter always returns a future:
    assert proxy.nested_object_property.get().inner == "nested_with_attr_marker.inner"


def test_traversable_cannot_mark_object_using_slots():
    with pytest.raises(Exception) as exc_info:
        pykka.traversable(NestedWithNoMarkerAndSlots())

    assert "cannot be used to mark an object using slots" in str(exc_info.value)



import logging
import sys
import threading
import uuid

from pykka import ActorDeadError, ActorRef, ActorRegistry, messages

__all__ = ["Actor"]

logger = logging.getLogger("pykka")


class Actor:
    """
    To create an actor:

    1. subclass one of the :class:`Actor` implementations:

       - :class:`~pykka.ThreadingActor`

    2. implement your methods, including :meth:`__init__`, as usual,
    3. call :meth:`Actor.start` on your actor class, passing the method any
       arguments for your constructor.

    To stop an actor, call :meth:`Actor.stop()` or :meth:`ActorRef.stop()`.

    For example::

        import pykka

        class MyActor(pykka.ThreadingActor):
            def __init__(self, my_arg=None):
                super().__init__()
                ... # My optional init code with access to start() arguments

            def on_start(self):
                ... # My optional setup code in same context as on_receive()

            def on_stop(self):
                ... # My optional cleanup code in same context as on_receive()

            def on_failure(self, exception_type, exception_value, traceback):
                ... # My optional cleanup code in same context as on_receive()

            def on_receive(self, message):
                ... # My optional message handling code for a plain actor

            def a_method(self, ...):
                ... # My regular method to be used through an ActorProxy

        my_actor_ref = MyActor.start(my_arg=...)
        my_actor_ref.stop()
    """

    @classmethod
    def start(cls, *args, **kwargs):
        """
        Start an actor and register it in the
        :class:`ActorRegistry <pykka.ActorRegistry>`.

        Any arguments passed to :meth:`start` will be passed on to the class
        constructor.

        Behind the scenes, the following is happening when you call
        :meth:`start`:

        1. The actor is created:

           1. :attr:`actor_urn` is initialized with the assigned URN.

           2. :attr:`actor_inbox` is initialized with a new actor inbox.

           3. :attr:`actor_ref` is initialized with a :class:`pykka.ActorRef`
              object for safely communicating with the actor.

           4. At this point, your :meth:`__init__()` code can run.

        2. The actor is registered in :class:`pykka.ActorRegistry`.

        3. The actor receive loop is started by the actor's associated
           thread/greenlet.

        :returns: a :class:`ActorRef` which can be used to access the actor in
            a safe manner
        """
        obj = cls(*args, **kwargs)
        assert obj.actor_ref is not None, (
            "Actor.__init__() have not been called. "
            "Did you forget to call super() in your override?"
        )
        ActorRegistry.register(obj.actor_ref)
        logger.debug(f"Starting {obj}")
        obj._start_actor_loop()
        return obj.actor_ref

    @staticmethod
    def _create_actor_inbox():
        """Internal method for implementors of new actor types."""
        raise NotImplementedError("Use a subclass of Actor")

    @staticmethod
    def _create_future():
        """Internal method for implementors of new actor types."""
        raise NotImplementedError("Use a subclass of Actor")

    def _start_actor_loop(self):
        """Internal method for implementors of new actor types."""
        raise NotImplementedError("Use a subclass of Actor")

    #: The actor URN string is a universally unique identifier for the actor.
    #: It may be used for looking up a specific actor using
    #: :meth:`ActorRegistry.get_by_urn`.
    actor_urn = None

    #: The actor's inbox. Use :meth:`ActorRef.tell`, :meth:`ActorRef.ask`, and
    #: friends to put messages in the inbox.
    actor_inbox = None

    #: The actor's :class:`ActorRef` instance.
    actor_ref = None

    #: A :class:`threading.Event` representing whether or not the actor should
    #: continue processing messages. Use :meth:`stop` to change it.
    actor_stopped = None

    def __init__(self, *args, **kwargs):
        """
        Your are free to override :meth:`__init__`, but you must call your
        superclass' :meth:`__init__` to ensure that fields :attr:`actor_urn`,
        :attr:`actor_inbox`, and :attr:`actor_ref` are initialized.

        You can use :func:`super`::

            super().__init__()

        Or call you superclass directly::

            pykka.ThreadingActor.__init__(self)

        :meth:`__init__` is called before the actor is started and registered
        in :class:`ActorRegistry <pykka.ActorRegistry>`.
        """
        self.actor_urn = uuid.uuid4().urn
        self.actor_inbox = self._create_actor_inbox()
        self.actor_stopped = threading.Event()

        self.actor_ref = ActorRef(self)

    def __str__(self):
        return f"{self.__class__.__name__} ({self.actor_urn})"

    def stop(self):
        """
        Stop the actor.

        It's equivalent to calling :meth:`ActorRef.stop` with ``block=False``.
        """
        self.actor_ref.tell(messages._ActorStop())

    def _stop(self):
        """
        Stops the actor immediately without processing the rest of the inbox.
        """
        ActorRegistry.unregister(self.actor_ref)
        self.actor_stopped.set()
        logger.debug(f"Stopped {self}")
        try:
            self.on_stop()
        except Exception:
            self._handle_failure(*sys.exc_info())

    def _actor_loop(self):
        """
        The actor's event loop.

        This is the method that will be executed by the thread or greenlet.
        """
        try:
            self.on_start()
        except Exception:
            self._handle_failure(*sys.exc_info())

        while not self.actor_stopped.is_set():
            envelope = self.actor_inbox.get()
            try:
                response = self._handle_receive(envelope.message)
                if envelope.reply_to is not None:
                    envelope.reply_to.set(response)
            except Exception:
                if envelope.reply_to is not None:
                    logger.info(
                        f"Exception returned from {self} to caller:",
                        exc_info=sys.exc_info(),
                    )
                    envelope.reply_to.set_exception()
                else:
                    self._handle_failure(*sys.exc_info())
                    try:
                        self.on_failure(*sys.exc_info())
                    except Exception:
                        self._handle_failure(*sys.exc_info())
            except BaseException:
                exception_value = sys.exc_info()[1]
                logger.debug(f"{exception_value!r} in {self}. Stopping all actors.")
                self._stop()
                ActorRegistry.stop_all()

        while not self.actor_inbox.empty():
            envelope = self.actor_inbox.get()
            if envelope.reply_to is not None:
                if isinstance(envelope.message, messages._ActorStop):
                    envelope.reply_to.set(None)
                else:
                    envelope.reply_to.set_exception(
                        exc_info=(
                            ActorDeadError,
                            ActorDeadError(
                                f"{self.actor_ref} stopped before "
                                f"handling the message"
                            ),
                            None,
                        )
                    )

    def on_start(self):
        """
        Hook for doing any setup that should be done *after* the actor is
        started, but *before* it starts processing messages.

        For :class:`ThreadingActor`, this method is executed in the actor's own
        thread, while :meth:`__init__` is executed in the thread that created
        the actor.

        If an exception is raised by this method the stack trace will be
        logged, and the actor will stop.
        """
        pass

    def on_stop(self):
        """
        Hook for doing any cleanup that should be done *after* the actor has
        processed the last message, and *before* the actor stops.

        This hook is *not* called when the actor stops because of an unhandled
        exception. In that case, the :meth:`on_failure` hook is called instead.

        For :class:`ThreadingActor` this method is executed in the actor's own
        thread, immediately before the thread exits.

        If an exception is raised by this method the stack trace will be
        logged, and the actor will stop.
        """
        pass

    def _handle_failure(self, exception_type, exception_value, traceback):
        """Logs unexpected failures, unregisters and stops the actor."""
        logger.error(
            f"Unhandled exception in {self}:",
            exc_info=(exception_type, exception_value, traceback),
        )
        ActorRegistry.unregister(self.actor_ref)
        self.actor_stopped.set()

    def on_failure(self, exception_type, exception_value, traceback):
        """
        Hook for doing any cleanup *after* an unhandled exception is raised,
        and *before* the actor stops.

        For :class:`ThreadingActor` this method is executed in the actor's own
        thread, immediately before the thread exits.

        The method's arguments are the relevant information from
        :func:`sys.exc_info`.

        If an exception is raised by this method the stack trace will be
        logged, and the actor will stop.
        """
        pass

    def _handle_receive(self, message):
        """Handles messages sent to the actor."""
        if isinstance(message, messages._ActorStop):
            return self._stop()
        if isinstance(message, messages.ProxyCall):
            callee = self._get_attribute_from_path(message.attr_path)
            return callee(*message.args, **message.kwargs)
        if isinstance(message, messages.ProxyGetAttr):
            attr = self._get_attribute_from_path(message.attr_path)
            return attr
        if isinstance(message, messages.ProxySetAttr):
            parent_attr = self._get_attribute_from_path(message.attr_path[:-1])
            attr_name = message.attr_path[-1]
            return setattr(parent_attr, attr_name, message.value)
        return self.on_receive(message)

    def on_receive(self, message):
        """
        May be implemented for the actor to handle regular non-proxy messages.

        :param message: the message to handle
        :type message: any

        :returns: anything that should be sent as a reply to the sender
        """
        logger.warning(f"Unexpected message received by {self}: {message}")

    def _get_attribute_from_path(self, attr_path):
        """
        Traverses the path and returns the attribute at the end of the path.
        """
        attr = self
        for attr_name in attr_path:
            attr = getattr(attr, attr_name)
        return attr

    def _introspect_attribute_from_path(self, attr_path):
        """Get attribute information from ``__dict__`` on the container."""
        if not attr_path:
            return self

        parent = self._get_attribute_from_path(attr_path[:-1])
        parent_attrs = self._introspect_attributes(parent)
        attr_name = attr_path[-1]

        try:
            return parent_attrs[attr_name]
        except KeyError:
            raise AttributeError(
                f"type object {parent.__class__.__name__!r} "
                f"has no attribute {attr_name!r}"
            )

    def _introspect_attributes(self, obj):
        """Combine ``__dict__`` from ``obj`` and all its superclasses."""
        result = {}
        for cls in reversed(obj.__class__.mro()):
            result.update(cls.__dict__)
        if hasattr(obj, "__dict__"):
            result.update(obj.__dict__)
        return result



"""
The :mod:`pykka.messages` module contains Pykka's own actor messages.

In general, you should not need to use any of these classes. However, they have
been made part of the public API so that certain optimizations can be done
without touching Pykka's internals.

An example is to combine :meth:`~pykka.ActorRef.ask` and :class:`ProxyCall`
to call a method on an actor without having to spend any resources on creating
a proxy object::

    reply = actor_ref.ask(
        ProxyCall(
            attr_path=['my_method'],
            args=['foo'],
            kwargs={'bar': 'baz'}
        )
    )

Another example is to use :meth:`~pykka.ActorRef.tell` instead of
:meth:`~pykka.ActorRef.ask` for the proxy method call, and thus avoid the
creation of a future for the return value if you don't need it.

It should be noted that these optimizations should only be necessary in very
special circumstances.
"""

from typing import Any, Dict, NamedTuple, Sequence, Tuple


class _ActorStop(NamedTuple):
    """Internal message."""

    pass


class ProxyCall(NamedTuple):
    """Message to ask the actor to call the method with the arguments."""

    #: List with the path from the actor to the method.
    attr_path: Sequence[str]

    #: List with positional arguments.
    args: Tuple[Any]

    #: Dict with keyword arguments.
    kwargs: Dict[str, Any]


class ProxyGetAttr(NamedTuple):
    """Message to ask the actor to return the value of the attribute."""

    #: List with the path from the actor to the attribute.
    attr_path: Sequence[str]


class ProxySetAttr(NamedTuple):
    """Message to ask the actor to set the attribute to the value."""

    #: List with the path from the actor to the attribute.
    attr_path: Sequence[str]

    #: The value to set the attribute to.
    value: Any



import collections
import logging
import threading
import time


class PykkaTestLogHandler(logging.Handler):
    def __init__(self, *args, **kwargs):
        self.lock = threading.RLock()
        with self.lock:
            self.events = collections.defaultdict(threading.Event)
            self.messages = {}
            self.reset()
        logging.Handler.__init__(self, *args, **kwargs)

    def emit(self, record):
        with self.lock:
            level = record.levelname.lower()
            self.messages[level].append(record)
            self.events[level].set()

    def reset(self):
        with self.lock:
            for level in ("debug", "info", "warning", "error", "critical"):
                self.events[level].clear()
                self.messages[level] = []

    def wait_for_message(self, level, num_messages=1, timeout=5):
        """Wait until at least ``num_messages`` log messages have been emitted
        to the given log level."""
        deadline = time.time() + timeout
        while time.time() < deadline:
            with self.lock:
                if len(self.messages[level]) >= num_messages:
                    return
                self.events[level].clear()
            self.events[level].wait(1)
        raise Exception(f"Timeout: Waited {timeout:d}s for log message")



import threading
from collections.abc import Sequence
from types import TracebackType
from typing import Any, Protocol

from pykka import ActorRef, Future
from pykka._envelope import Envelope

class ActorInbox(Protocol):
    def put(self, envelope: Envelope) -> None: ...
    def get(self) -> Envelope: ...

class Actor:
    @classmethod
    def start(cls, *args: Any, **kwargs: Any) -> ActorRef: ...
    @staticmethod
    def _create_actor_inbox() -> ActorInbox: ...
    @staticmethod
    def _create_future() -> Future: ...
    def _start_actor_loop(self) -> None: ...
    actor_urn: str
    actor_inbox: ActorInbox
    actor_ref: ActorRef
    actor_stopped: threading.Event
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...
    def stop(self) -> None: ...
    def _stop(self) -> None: ...
    def _actor_loop(self) -> None: ...
    def on_start(self) -> None: ...
    def on_stop(self) -> None: ...
    def _handle_failure(
        self,
        exception_type: type[BaseException],
        exception_value: BaseException,
        traceback: TracebackType,
    ) -> None: ...
    def on_failure(
        self,
        exception_type: type[BaseException],
        exception_value: BaseException,
        traceback: TracebackType,
    ) -> None: ...
    def _handle_receive(self, message: Any) -> Any: ...
    def on_receive(self, message: Any) -> Any: ...
    def _get_attribute_from_path(self, attr_path: Sequence[str]) -> Any: ...
    def _introspect_attribute_from_path(
        self, attr_path: Sequence[str]
    ) -> dict[str, Any]: ...
    def _introspect_attributes(self, obj: Any) -> dict[str, Any]: ...



import pytest


@pytest.fixture
def actor_class(runtime):
    class ActorWithProperties(runtime.actor_class):
        an_attr = "an_attr"
        _private_attr = "secret"

        @property
        def a_ro_property(self):
            return "a_ro_property"

        _a_rw_property = "a_rw_property"

        @property
        def a_rw_property(self):
            return self._a_rw_property

        @a_rw_property.setter
        def a_rw_property(self, value):
            self._a_rw_property = value

    return ActorWithProperties


@pytest.fixture
def proxy(actor_class):
    proxy = actor_class.start().proxy()
    yield proxy
    proxy.stop()


def test_attr_can_be_read_using_get_postfix(proxy):
    assert proxy.an_attr.get() == "an_attr"


def test_attr_can_be_set_using_assignment(proxy):
    assert proxy.an_attr.get() == "an_attr"

    proxy.an_attr = "an_attr_2"

    assert proxy.an_attr.get() == "an_attr_2"


def test_private_attr_access_raises_exception(proxy):
    with pytest.raises(AttributeError) as exc_info:
        proxy._private_attr.get()

    assert "has no attribute '_private_attr'" in str(exc_info.value)


def test_missing_attr_access_raises_exception(proxy):
    with pytest.raises(AttributeError) as exc_info:
        proxy.missing_attr.get()

    assert "has no attribute 'missing_attr'" in str(exc_info.value)


def test_property_can_be_read_using_get_postfix(proxy):
    assert proxy.a_ro_property.get() == "a_ro_property"
    assert proxy.a_rw_property.get() == "a_rw_property"


def test_property_can_be_set_using_assignment(proxy):
    proxy.a_rw_property = "a_rw_property_2"

    assert proxy.a_rw_property.get() == "a_rw_property_2"


def test_read_only_property_cannot_be_set(proxy):
    with pytest.raises(AttributeError):
        proxy.a_ro_property = "a_ro_property_2"


def test_property_is_not_accessed_when_creating_proxy(runtime):
    class ExpensiveSideEffectActor(runtime.actor_class):
        @property
        def a_property(self):
            # Imagine code with side effects or heavy resource usage here
            raise Exception("Proxy creation accessed property")

    actor_ref = ExpensiveSideEffectActor.start()
    try:
        actor_ref.proxy()
    finally:
        actor_ref.stop()



#!/usr/bin/env python3

import threading
import time

import pykka


class AnActor(pykka.ThreadingActor):
    field = "this is the value of AnActor.field"

    def proc(self):
        log("this was printed by AnActor.proc()")

    def func(self):
        time.sleep(0.5)  # Block a bit to make it realistic
        return "this was returned by AnActor.func() after a delay"


def log(msg):
    thread_name = threading.current_thread().name
    print(f"{thread_name}: {msg}")


if __name__ == "__main__":
    actor = AnActor.start().proxy()
    for _ in range(3):
        # Method with side effect
        log("calling AnActor.proc() ...")
        actor.proc()

        # Method with return value
        log("calling AnActor.func() ...")
        result = actor.func()  # Does not block, returns a future
        log("printing result ... (blocking)")
        log(result.get())  # Blocks until ready

        # Field reading
        log("reading AnActor.field ...")
        result = actor.field  # Does not block, returns a future
        log("printing result ... (blocking)")
        log(result.get())  # Blocks until ready

        # Field writing
        log("writing AnActor.field ...")
        actor.field = "new value"  # Assignment does not block
        result = actor.field  # Does not block, returns a future
        log("printing new field value ... (blocking)")
        log(result.get())  # Blocks until ready
    actor.stop()



import uuid

import pytest

from pykka import ActorDeadError, ActorRegistry

pytestmark = pytest.mark.usefixtures("stop_all")


@pytest.fixture(scope="module")
def actor_class(runtime):
    class ActorA(runtime.actor_class):
        def __init__(self, events):
            super().__init__()
            self.events = events

        def on_start(self):
            self.events.on_start_was_called.set()
            if ActorRegistry.get_by_urn(self.actor_urn) is not None:
                self.events.actor_registered_before_on_start_was_called.set()

        def on_stop(self):
            self.events.on_stop_was_called.set()

        def on_failure(self, *args):
            self.events.on_failure_was_called.set()

        def on_receive(self, message):
            if message.get("command") == "raise exception":
                raise Exception("foo")
            elif message.get("command") == "raise base exception":
                raise BaseException()
            elif message.get("command") == "stop twice":
                self.stop()
                self.stop()
            elif message.get("command") == "message self then stop":
                self.actor_ref.tell({"command": "greetings"})
                self.stop()
            elif message.get("command") == "greetings":
                self.events.greetings_was_received.set()
            elif message.get("command") == "callback":
                message["callback"]()
            else:
                super().on_receive(message)

    return ActorA


@pytest.fixture
def actor_ref(actor_class, events):
    ref = actor_class.start(events)
    yield ref
    ref.stop()


@pytest.fixture(scope="module")
def early_stopping_actor_class(runtime):
    class EarlyStoppingActor(runtime.actor_class):
        def __init__(self, events):
            super().__init__()
            self.events = events

        def on_start(self):
            self.stop()

        def on_stop(self):
            self.events.on_stop_was_called.set()

    return EarlyStoppingActor


def test_messages_left_in_queue_after_actor_stops_receive_an_error(runtime, actor_ref):
    event = runtime.event_class()

    actor_ref.tell({"command": "callback", "callback": event.wait})
    actor_ref.stop(block=False)
    response = actor_ref.ask({"command": "irrelevant"}, block=False)
    event.set()

    with pytest.raises(ActorDeadError):
        response.get(timeout=0.5)


def test_stop_requests_left_in_queue_after_actor_stops_are_handled(runtime, actor_ref):
    event = runtime.event_class()

    actor_ref.tell({"command": "callback", "callback": event.wait})
    actor_ref.stop(block=False)
    response = actor_ref.stop(block=False)
    event.set()

    response.get(timeout=0.5)


def test_actor_has_an_uuid4_based_urn(actor_ref):
    assert uuid.UUID(actor_ref.actor_urn).version == 4


def test_actor_has_unique_uuid(actor_class, events):
    actors = [actor_class.start(events) for _ in range(3)]

    assert actors[0].actor_urn != actors[1].actor_urn
    assert actors[1].actor_urn != actors[2].actor_urn
    assert actors[2].actor_urn != actors[0].actor_urn


def test_str_on_raw_actor_contains_actor_class_name(actor_class, events):
    unstarted_actor = actor_class(events)

    assert "ActorA" in str(unstarted_actor)


def test_str_on_raw_actor_contains_actor_urn(actor_class, events):
    unstarted_actor = actor_class(events)

    assert unstarted_actor.actor_urn in str(unstarted_actor)


def test_init_can_be_called_with_arbitrary_arguments(runtime):
    runtime.actor_class(1, 2, 3, foo="bar")


def test_on_start_is_called_before_first_message_is_processed(actor_ref, events):
    events.on_start_was_called.wait(5)
    assert events.on_start_was_called.is_set()


def test_on_start_is_called_after_the_actor_is_registered(actor_ref, events):
    # NOTE: If the actor is registered after the actor is started, this
    # test may still occasionally pass, as it is dependant on the exact
    # timing of events. When the actor is first registered and then
    # started, this test should always pass.
    events.on_start_was_called.wait(5)
    assert events.on_start_was_called.is_set()

    events.actor_registered_before_on_start_was_called.wait(0.1)
    assert events.actor_registered_before_on_start_was_called.is_set()


def test_on_start_can_stop_actor_before_receive_loop_is_started(
    early_stopping_actor_class, events
):
    # NOTE: This test will pass even if the actor is allowed to start the
    # receive loop, but it will cause the test suite to hang, as the actor
    # thread is blocking on receiving messages to the actor inbox forever.
    # If one made this test specifically for ThreadingActor, one could add
    # an assertFalse(actor_thread.is_alive()), which would cause the test
    # to fail properly.
    actor_ref = early_stopping_actor_class.start(events)

    events.on_stop_was_called.wait(5)
    assert events.on_stop_was_called.is_set()
    assert not actor_ref.is_alive()


def test_on_start_failure_causes_actor_to_stop(early_failing_actor_class, events):
    # Actor should not be alive if on_start fails.

    actor_ref = early_failing_actor_class.start(events)
    events.on_start_was_called.wait(5)

    actor_ref.actor_stopped.wait(5)
    assert not actor_ref.is_alive()


def test_on_stop_is_called_when_actor_is_stopped(actor_ref, events):
    assert not events.on_stop_was_called.is_set()

    actor_ref.stop()

    events.on_stop_was_called.wait(5)
    assert events.on_stop_was_called.is_set()


def test_on_stop_failure_causes_actor_to_stop(late_failing_actor_class, events):
    actor_ref = late_failing_actor_class.start(events)

    events.on_stop_was_called.wait(5)
    assert not actor_ref.is_alive()


def test_on_failure_is_called_when_exception_cannot_be_returned(actor_ref, events):
    assert not events.on_failure_was_called.is_set()

    actor_ref.tell({"command": "raise exception"})

    events.on_failure_was_called.wait(5)
    assert events.on_failure_was_called.is_set()
    assert not events.on_stop_was_called.is_set()


def test_on_failure_failure_causes_actor_to_stop(
    failing_on_failure_actor_class, events
):
    actor_ref = failing_on_failure_actor_class.start(events)

    actor_ref.tell({"command": "raise exception"})

    events.on_failure_was_called.wait(5)
    assert not actor_ref.is_alive()


def test_actor_is_stopped_when_unhandled_exceptions_are_raised(actor_ref, events):
    assert not events.on_failure_was_called.is_set()

    actor_ref.tell({"command": "raise exception"})

    events.on_failure_was_called.wait(5)
    assert events.on_failure_was_called.is_set()
    assert len(ActorRegistry.get_all()) == 0


def test_all_actors_are_stopped_on_base_exception(events, actor_ref):
    assert len(ActorRegistry.get_all()) == 1
    assert not events.on_stop_was_called.is_set()

    actor_ref.tell({"command": "raise base exception"})

    events.on_stop_was_called.wait(5)
    assert events.on_stop_was_called.is_set()
    assert len(ActorRegistry.get_all()) == 0

    events.on_stop_was_called.wait(5)
    assert events.on_stop_was_called.is_set()
    assert len(ActorRegistry.get_all()) == 0


def test_actor_can_call_stop_on_self_multiple_times(actor_ref):
    actor_ref.ask({"command": "stop twice"})


def test_actor_processes_all_messages_before_stop_on_self_stops_it(actor_ref, events):
    actor_ref.ask({"command": "message self then stop"})

    events.greetings_was_received.wait(5)
    assert events.greetings_was_received.is_set()

    events.on_stop_was_called.wait(5)
    assert len(ActorRegistry.get_all()) == 0



import pytest

from pykka import ActorDeadError, Timeout


@pytest.fixture(scope="module")
def actor_class(runtime):
    class ActorA(runtime.actor_class):
        received_messages = None

        def __init__(self, received_message):
            super(runtime.actor_class, self).__init__()
            self.received_message = received_message

        def on_receive(self, message):
            if isinstance(message, dict) and message.get("command") == "ping":
                runtime.sleep_func(0.01)
                return "pong"
            else:
                self.received_message.set(message)

    return ActorA


@pytest.fixture
def received_messages(runtime):
    return runtime.future_class()


@pytest.fixture
def actor_ref(actor_class, received_messages):
    ref = actor_class.start(received_messages)
    yield ref
    ref.stop()


def test_repr_is_wrapped_in_lt_and_gt(actor_ref):
    result = repr(actor_ref)

    assert result.startswith("<")
    assert result.endswith(">")


def test_repr_reveals_that_this_is_a_ref(actor_ref):
    assert "ActorRef" in repr(actor_ref)


def test_repr_contains_actor_class_name(actor_ref):
    assert "ActorA" in repr(actor_ref)


def test_repr_contains_actor_urn(actor_ref):
    assert actor_ref.actor_urn in repr(actor_ref)


def test_str_contains_actor_class_name(actor_ref):
    assert "ActorA" in str(actor_ref)


def test_str_contains_actor_urn(actor_ref):
    assert actor_ref.actor_urn in str(actor_ref)


def test_is_alive_returns_true_for_running_actor(actor_ref):
    assert actor_ref.is_alive()


def test_is_alive_returns_false_for_dead_actor(actor_ref):
    actor_ref.stop()

    assert not actor_ref.is_alive()


def test_stop_returns_true_if_actor_is_stopped(actor_ref):
    assert actor_ref.stop()


def test_stop_does_not_stop_already_dead_actor(actor_ref):
    assert actor_ref.stop()
    assert not actor_ref.stop()


def test_tell_delivers_message_to_actors_custom_on_receive(
    actor_ref, received_messages
):
    actor_ref.tell({"command": "a custom message"})

    assert received_messages.get(timeout=1) == {"command": "a custom message"}


@pytest.mark.parametrize(
    "message",
    [
        123,
        123.456,
        {"a": "dict"},
        ("a", "tuple"),
        ["a", "list"],
        Exception("an exception"),
    ],
)
def test_tell_accepts_any_object_as_the_message(actor_ref, message, received_messages):
    actor_ref.tell(message)

    assert received_messages.get(timeout=1) == message


def test_tell_fails_if_actor_is_stopped(actor_ref):
    actor_ref.stop()

    with pytest.raises(ActorDeadError) as exc_info:
        actor_ref.tell({"command": "a custom message"})

    assert str(exc_info.value) == f"{actor_ref} not found"


def test_ask_blocks_until_response_arrives(actor_ref):
    result = actor_ref.ask({"command": "ping"})

    assert result == "pong"


def test_ask_can_timeout_if_blocked_too_long(actor_ref):
    with pytest.raises(Timeout):
        actor_ref.ask({"command": "ping"}, timeout=0)


def test_ask_can_return_future_instead_of_blocking(actor_ref):
    future = actor_ref.ask({"command": "ping"}, block=False)

    assert future.get() == "pong"


def test_ask_fails_if_actor_is_stopped(actor_ref):
    actor_ref.stop()

    with pytest.raises(ActorDeadError) as exc_info:
        actor_ref.ask({"command": "ping"})

    assert str(exc_info.value) == f"{actor_ref} not found"


def test_ask_nonblocking_fails_future_if_actor_is_stopped(actor_ref):
    actor_ref.stop()
    future = actor_ref.ask({"command": "ping"}, block=False)

    with pytest.raises(ActorDeadError) as exc_info:
        future.get()

    assert str(exc_info.value) == f"{actor_ref} not found"



from collections.abc import Callable, Sequence
from typing import Any, NamedTuple, TypeVar

from pykka import Actor, ActorRef, Future

AttrPath = Sequence[str]

class AttrInfo(NamedTuple):
    callable: bool
    traversable: bool

class ActorProxy:
    actor_ref: ActorRef
    _actor: Actor
    _attr_path: AttrPath
    _known_attrs: dict[AttrPath, AttrInfo]
    _actor_proxies: dict[AttrPath, ActorProxy]
    _callable_proxies: dict[AttrPath, CallableProxy]
    def __init__(
        self, actor_ref: ActorRef, attr_path: AttrPath | None = ...
    ) -> None: ...
    def _introspect_attributes(self) -> dict[AttrPath, AttrInfo]: ...
    def _is_exposable_attribute(self, attr_name: str) -> bool: ...
    def _is_self_proxy(self, attr: Any) -> bool: ...
    def _is_callable_attribute(self, attr: Any) -> bool: ...
    def _is_traversable_attribute(self, attr: Any) -> bool: ...
    def __eq__(self, other: object) -> bool: ...
    def __hash__(self) -> int: ...
    def __dir__(self) -> dict[str, Any]: ...
    def __getattr__(self, name: str) -> CallableProxy | ActorProxy | Future[Any]: ...
    def __setattr__(self, name: str, value: Any) -> None: ...

class CallableProxy:
    actor_ref: ActorRef
    _attr_path: AttrPath
    def __init__(self, actor_ref: ActorRef, attr_path: AttrPath) -> None: ...
    def __call__(self, *args: tuple[Any], **kwargs: dict[str, Any]) -> Future[Any]: ...
    def defer(self, *args: tuple[Any], **kwargs: dict[str, Any]) -> None: ...

FuncType = Callable[..., Any]
_F = TypeVar("_F", bound=FuncType)

def traversable(obj: _F) -> _F: ...



import threading
from typing import Any, overload

from typing_extensions import Literal  # Py38+: Available in ``typing``

from pykka import ActorProxy, Future
from pykka._actor import Actor, ActorInbox

class ActorRef:
    _actor: Actor
    actor_class: type[Actor]
    actor_urn: str
    actor_inbox: ActorInbox
    actor_stopped: threading.Event
    def __init__(self, actor: Actor) -> None: ...
    def is_alive(self) -> bool: ...
    def tell(self, message: Any) -> None: ...
    @overload
    def ask(
        self,
        message: Any,
        block: Literal[False],
        timeout: float | None = ...,
    ) -> Future[Any]: ...
    @overload  # noqa: Allow redefinition
    def ask(
        self,
        message: Any,
        block: Literal[True],
        timeout: float | None = ...,
    ) -> Any: ...
    @overload  # noqa: Allow redefinition
    def ask(
        self, message: Any, block: bool = ..., timeout: float | None = ...
    ) -> Future[Any] | Any: ...
    @overload
    def stop(self, block: Literal[True], timeout: float | None) -> bool: ...
    @overload  # noqa: Allow redefinition
    def stop(
        self, block: Literal[False], timeout: float | None = ...
    ) -> Future[bool]: ...
    @overload  # noqa: Allow redefinition
    def stop(self, block: bool, timeout: float | None = ...) -> Future[bool] | bool: ...
    def proxy(self) -> ActorProxy: ...



import logging

import pytest

pytestmark = pytest.mark.usefixtures("stop_all")


@pytest.fixture(scope="module")
def actor_class(runtime):
    class ActorA(runtime.actor_class):
        def __init__(self, events):
            super().__init__()
            self.events = events

        def on_stop(self):
            self.events.on_stop_was_called.set()

        def on_failure(self, exception_type, exception_value, traceback):
            self.events.on_failure_was_called.set()

        def on_receive(self, message):
            if message.get("command") == "raise exception":
                return self.raise_exception()
            elif message.get("command") == "raise base exception":
                raise BaseException()
            else:
                super().on_receive(message)

        def raise_exception(self):
            raise Exception("foo")

    return ActorA


@pytest.fixture
def actor_ref(actor_class, events):
    ref = actor_class.start(events)
    yield ref
    ref.stop()


def test_null_handler_is_added_to_avoid_warnings():
    logger = logging.getLogger("pykka")
    handler_names = [h.__class__.__name__ for h in logger.handlers]

    assert "NullHandler" in handler_names


def test_unexpected_messages_are_logged(actor_ref, log_handler):
    actor_ref.ask({"unhandled": "message"})

    log_handler.wait_for_message("warning")
    with log_handler.lock:
        assert len(log_handler.messages["warning"]) == 1
        log_record = log_handler.messages["warning"][0]

    assert log_record.getMessage().split(": ")[0] == (
        f"Unexpected message received by {actor_ref}"
    )


def test_exception_is_logged_when_returned_to_caller(actor_ref, log_handler):
    with pytest.raises(Exception):
        actor_ref.proxy().raise_exception().get()

    log_handler.wait_for_message("info")
    with log_handler.lock:
        assert len(log_handler.messages["info"]) == 1
        log_record = log_handler.messages["info"][0]

    assert log_record.getMessage() == (
        f"Exception returned from {actor_ref} to caller:"
    )
    assert log_record.exc_info[0] == Exception
    assert str(log_record.exc_info[1]) == "foo"


def test_exception_is_logged_when_not_reply_requested(actor_ref, events, log_handler):
    events.on_failure_was_called.clear()
    actor_ref.tell({"command": "raise exception"})

    events.on_failure_was_called.wait(5)
    assert events.on_failure_was_called.is_set()

    log_handler.wait_for_message("error")
    with log_handler.lock:
        assert len(log_handler.messages["error"]) == 1
        log_record = log_handler.messages["error"][0]

    assert log_record.getMessage() == f"Unhandled exception in {actor_ref}:"
    assert log_record.exc_info[0] == Exception
    assert str(log_record.exc_info[1]) == "foo"


def test_base_exception_is_logged(actor_ref, events, log_handler):
    log_handler.reset()
    events.on_stop_was_called.clear()
    actor_ref.tell({"command": "raise base exception"})

    events.on_stop_was_called.wait(5)
    assert events.on_stop_was_called.is_set()

    log_handler.wait_for_message("debug", num_messages=3)
    with log_handler.lock:
        assert len(log_handler.messages["debug"]) == 3
        log_record = log_handler.messages["debug"][0]

    assert log_record.getMessage() == (
        f"BaseException() in {actor_ref}. Stopping all actors."
    )


def test_exception_in_on_start_is_logged(
    early_failing_actor_class, events, log_handler
):
    log_handler.reset()
    actor_ref = early_failing_actor_class.start(events)
    events.on_start_was_called.wait(5)

    log_handler.wait_for_message("error")
    with log_handler.lock:
        assert len(log_handler.messages["error"]) == 1
        log_record = log_handler.messages["error"][0]

    assert log_record.getMessage() == f"Unhandled exception in {actor_ref}:"


def test_exception_in_on_stop_is_logged(late_failing_actor_class, events, log_handler):
    log_handler.reset()
    actor_ref = late_failing_actor_class.start(events)
    events.on_stop_was_called.wait(5)

    log_handler.wait_for_message("error")
    with log_handler.lock:
        assert len(log_handler.messages["error"]) == 1
        log_record = log_handler.messages["error"][0]

    assert log_record.getMessage() == f"Unhandled exception in {actor_ref}:"


def test_exception_in_on_failure_is_logged(
    failing_on_failure_actor_class, events, log_handler
):
    log_handler.reset()
    actor_ref = failing_on_failure_actor_class.start(events)
    actor_ref.tell({"command": "raise exception"})
    events.on_failure_was_called.wait(5)

    log_handler.wait_for_message("error", num_messages=2)
    with log_handler.lock:
        assert len(log_handler.messages["error"]) == 2
        log_record = log_handler.messages["error"][0]

    assert log_record.getMessage() == f"Unhandled exception in {actor_ref}:"



import pytest
from producer import ProducerActor


@pytest.fixture
def consumer_mock(mocker):
    yield mocker.Mock()


@pytest.fixture
def producer(consumer_mock):
    # Step 1: The actor under test is wired up with
    # its dependencies and is started.
    proxy = ProducerActor.start(consumer_mock).proxy()

    yield proxy

    # Step 4: The actor is stopped to clean up before the next test.
    proxy.stop()


def test_producer_actor(consumer_mock, producer):
    # Step 2: Interact with the actor.
    # We call .get() on the last future returned by the actor to wait
    # for the actor to process all messages before asserting anything.
    producer.produce().get()

    # Step 3: Assert that the return values or actor state is as expected.
    consumer_mock.consume.assert_called_once_with({"item": 1, "new": True})



from pykka.messages import ProxyCall, ProxyGetAttr, ProxySetAttr, _ActorStop


def test_actor_stop():
    message = _ActorStop()

    assert isinstance(message, _ActorStop)


def test_proxy_call():
    message = ProxyCall(attr_path=["nested", "method"], args=[1], kwargs={"a": "b"})

    assert isinstance(message, ProxyCall)
    assert message.attr_path == ["nested", "method"]
    assert message.args == [1]
    assert message.kwargs == {"a": "b"}


def test_proxy_get_attr():
    message = ProxyGetAttr(attr_path=["nested", "attr"])

    assert isinstance(message, ProxyGetAttr)
    assert message.attr_path == ["nested", "attr"]


def test_proxy_set_attr():
    message = ProxySetAttr(attr_path=["nested", "attr"], value="abcdef")

    assert isinstance(message, ProxySetAttr)
    assert message.attr_path == ["nested", "attr"]
    assert message.value == "abcdef"



import pytest


@pytest.fixture(scope="module")
def actor_class(runtime):
    class ActorA(runtime.actor_class):
        def add_method(self, name):
            setattr(self, name, lambda: "returned by " + name)

        def use_foo_through_self_proxy(self):
            return self.actor_ref.proxy().foo()

    return ActorA


@pytest.fixture
def proxy(actor_class):
    proxy = actor_class.start().proxy()
    yield proxy
    proxy.stop()


def test_can_call_method_that_was_added_at_runtime(proxy):
    # We need to .get() after .add_method() to be sure that the method has
    # been added before we try to use it through the proxy.
    proxy.add_method("foo").get()

    assert proxy.foo().get() == "returned by foo"


def test_can_proxy_itself_and_use_attrs_added_at_runtime(proxy):
    # We don't need to .get() after .add_method() here, because the actor
    # will process the .add_method() call before processing the
    # .use_foo_through_self_proxy() call, which again will use the new
    # method, .foo().
    proxy.add_method("foo")

    outer_future = proxy.use_foo_through_self_proxy()
    inner_future = outer_future.get(timeout=1)
    result = inner_future.get(timeout=1)

    assert result == "returned by foo"



#!/usr/bin/env python3

import pykka


class Adder(pykka.ThreadingActor):
    def add_one(self, i):
        print(f"{self} is increasing {i}")
        return i + 1


class Bookkeeper(pykka.ThreadingActor):
    def __init__(self, adder):
        super().__init__()
        self.adder = adder

    def count_to(self, target):
        i = 0
        while i < target:
            i = self.adder.add_one(i).get()
            print(f"{self} got {i} back")


if __name__ == "__main__":
    adder = Adder.start().proxy()
    bookkeeper = Bookkeeper.start(adder).proxy()
    bookkeeper.count_to(10).get()
    pykka.ActorRegistry.stop_all()



#!/usr/bin/env python3

"""
Resolve a bunch of IP addresses using a pool of resolver actors.

Based on example contributed by Kristian Klette <klette@klette.us>.

Either run without arguments:

    ./resolver.py

Or specify pool size and IPs to resolve:

    ./resolver.py 3 193.35.52.{1,2,3,4,5,6,7,8,9}
"""

import pprint
import socket
import sys

import pykka


class Resolver(pykka.ThreadingActor):
    def resolve(self, ip):
        try:
            info = socket.gethostbyaddr(ip)
            print(f"Finished resolving {ip}")
            return info[0]
        except Exception:
            print(f"Failed resolving {ip}")
            return None


def run(pool_size, *ips):
    # Start resolvers
    resolvers = [Resolver.start().proxy() for _ in range(pool_size)]

    # Distribute work by mapping IPs to resolvers (not blocking)
    hosts = []
    for i, ip in enumerate(ips):
        hosts.append(resolvers[i % len(resolvers)].resolve(ip))

    # Gather results (blocking)
    ip_to_host = zip(ips, pykka.get_all(hosts))
    pprint.pprint(list(ip_to_host))

    # Clean up
    pykka.ActorRegistry.stop_all()


if __name__ == "__main__":
    if len(sys.argv[1:]) >= 2:
        run(int(sys.argv[1]), *sys.argv[2:])
    else:
        ips = [f"193.35.52.{i}" for i in range(1, 50)]
        run(10, *ips)



import pytest


@pytest.fixture(scope="module")
def actor_class(runtime):
    class ActorA(runtime.actor_class):
        cat = "dog"

        def __init__(self, events):
            super().__init__()
            self.events = events

        def on_stop(self):
            self.events.on_stop_was_called.set()

        def on_failure(self, *args):
            self.events.on_failure_was_called.set()

        def functional_hello(self, s):
            return f"Hello, {s}!"

        def set_cat(self, s):
            self.cat = s

        def raise_exception(self):
            raise Exception("boom!")

        def talk_with_self(self):
            return self.actor_ref.proxy().functional_hello("from the future")

    return ActorA


@pytest.fixture
def proxy(actor_class, events):
    proxy = actor_class.start(events).proxy()
    yield proxy
    proxy.stop()


def test_functional_method_call_returns_correct_value(proxy):
    assert proxy.functional_hello("world").get() == "Hello, world!"
    assert proxy.functional_hello("moon").get() == "Hello, moon!"


def test_side_effect_of_method_call_is_observable(proxy):
    assert proxy.cat.get() == "dog"

    future = proxy.set_cat("eagle")

    assert future.get() is None
    assert proxy.cat.get() == "eagle"


def test_side_effect_of_deferred_method_call_is_observable(proxy):
    assert proxy.cat.get() == "dog"

    result = proxy.set_cat.defer("eagle")

    assert result is None
    assert proxy.cat.get() == "eagle"


def test_exception_in_method_reraised_by_future(proxy, events):
    assert not events.on_failure_was_called.is_set()

    future = proxy.raise_exception()
    with pytest.raises(Exception) as exc_info:
        future.get()

    assert str(exc_info.value) == "boom!"
    assert not events.on_failure_was_called.is_set()


def test_exception_in_deferred_method_call_triggers_on_failure(proxy, events):
    assert not events.on_failure_was_called.is_set()

    result = proxy.raise_exception.defer()

    assert result is None
    events.on_failure_was_called.wait(5)
    assert events.on_failure_was_called.is_set()
    assert not events.on_stop_was_called.is_set()


def test_call_to_unknown_method_raises_attribute_error(proxy):
    with pytest.raises(AttributeError) as exc_info:
        proxy.unknown_method()

    result = str(exc_info.value)

    assert result.startswith("<ActorProxy for ActorA")
    assert result.endswith("has no attribute 'unknown_method'")


def test_deferred_call_to_unknown_method_raises_attribute_error(proxy):
    with pytest.raises(AttributeError) as exc_info:
        proxy.unknown_method.defer()

    result = str(exc_info.value)

    assert result.startswith("<ActorProxy for ActorA")
    assert result.endswith("has no attribute 'unknown_method'")


def test_can_proxy_itself_for_offloading_work_to_the_future(proxy):
    outer_future = proxy.talk_with_self()
    inner_future = outer_future.get(timeout=1)

    result = inner_future.get(timeout=1)

    assert result == "Hello, from the future!"



import threading
from typing import Any, ClassVar, overload

from typing_extensions import Literal  # Py38+: Available in ``typing``

from pykka import Actor, ActorRef, Future

class ActorRegistry:
    _actor_refs: ClassVar[list[ActorRef]]
    _actor_refs_lock: ClassVar[threading.RLock]
    @classmethod
    def broadcast(
        cls, message: Any, target_class: type[Actor] | None = ...
    ) -> None: ...
    @classmethod
    def get_all(cls) -> list[ActorRef]: ...
    @classmethod
    def get_by_class(cls, actor_class: type[Actor]) -> list[ActorRef]: ...
    @classmethod
    def get_by_class_name(cls, actor_class_name: str) -> list[ActorRef]: ...
    @classmethod
    def get_by_urn(cls, actor_urn: str) -> ActorRef | None: ...
    @classmethod
    def register(cls, actor_ref: ActorRef) -> None: ...
    @overload
    @classmethod
    def stop_all(
        cls, block: Literal[True], timeout: float | None = ...
    ) -> list[bool]: ...
    @overload  # noqa: Allow redefinition
    @classmethod
    def stop_all(
        cls, block: Literal[False], timeout: float | None = ...
    ) -> list[Future[bool]]: ...
    @overload  # noqa: Allow redefinition
    @classmethod
    def stop_all(
        cls, block: bool = ..., timeout: float | None = ...
    ) -> list[bool] | list[Future[bool]] | list[bool | Future[bool]]: ...
    @classmethod
    def unregister(cls, actor_ref: ActorRef) -> None: ...



__all__ = ["ActorDeadError", "Timeout"]


class ActorDeadError(Exception):
    """Exception raised when trying to use a dead or unavailable actor."""

    pass


class Timeout(Exception):
    """Exception raised at future timeout."""

    pass



#!/usr/bin/env python3

import pykka

GetMessages = object()


class PlainActor(pykka.ThreadingActor):
    def __init__(self):
        super().__init__()
        self.stored_messages = []

    def on_receive(self, message):
        if message is GetMessages:
            return self.stored_messages
        else:
            self.stored_messages.append(message)


if __name__ == "__main__":
    actor = PlainActor.start()
    actor.tell({"no": "Norway", "se": "Sweden"})
    actor.tell({"a": 3, "b": 4, "c": 5})
    print(actor.ask(GetMessages))
    actor.stop()
